/*
 *      This plugin should reconstruct object boundaries from a particular pointer.
 *      It will parse all xrefs to that pointer in the particular subfunction.      
 *  
 *  
 *    -  first working version at the 21st of March
 *    +  23.03.01:   
 *    +  24.03.01:   Added MAX_STEPS definition and a special maxsteps variable
 *                   into all code-parsing while loops to prevent the plugin from
 *                   seemingly hanging when code cross references form a kind of
 *                   infinite loop
 *    +  02.04.01:   Seeing what _Stone needs for his work, started to add both
 *                   register argument passing into the recursion algo as well
 *                   as area parsing and using regs...
 *       04.04.01:   Major re-write of the main run() function, started rewrite of
 *                   ArgRec() as consequence
 *       08.04.01:   Rewrite of ArgRec() and run() is completed to a degree that
 *                   non-call-recursive reconstruction works with complete register
 *                   tracing. Work on __cdecl and __fastcall recursion starts.            
 *                   The entire Do_Recursion() function will be killed. Furthermore,
 *                   the __cdecl-recursion will work without full direct call tree 
 *                   reconstruction (greets to the_owl at this point who helped with
 *                   the old code, got rid of it now but thx anyways :)
 *    + KNOWN BUG:   Due to lack of JMP tracing, incorrect structure members can
 *                   be created under certain conditions
 *       11.06.01:   Finally continued this :-) Changed ArgRec() to construct member
 *                   names from StrucName_offset instead of all members being called
 *                   anonymous_x. For this I moved StrucName to be a global variable.
 *                   (Yes I know that is bad style). This should make spath happier :)
 *       27.06.01:   I need to get some discipline when working on this :-o
 *                   Added Push-to-call association for __cdecl recursion.
 *                   Fucking shit, for some reasion SICE fails to load symbols for   
 *                   files generated with 4.16 SDK ??
 *                   Fixed bug in the LEA & MOV handling, now correctly aborts when
 *                   the register being traced is overwritten
 *                   Fixed another bug in the tracing engine - it would incorrectly
 *                   assume that a reference like [ecx+0x0C] would be an access to
 *                   a stack variable if that one was located at [ebp+0x0C]. Worked
 *                   around it by assuming that only references containing ESP or EBP
 *                   are valid.
 *                   Added very simple __cdecl recursion parsing - still need to add
 *                   stub-jump parsing and dealing with infinite looping
 *                   The lack of true code block tracing is somewhat annoying, but
 *                   due to the rapid-fire use now it can at least be fixed manually.
 *       28.06.01:   Adding vtable parsing today I think -- it should be fairly
 *                   easy
 *       01.07.01:   Still not done adding the vtable. Works nice so far, but I should
 *                   introduce a maximum recursion depth.
 *                   vtable parsing added at the price of unclean code between run()
 *                   and parsevtable. Has to be cleaned up ASAP.
 *                   Bug in __cdecl recursion: push [ebp+var] will not be recognized.
 *                   Not trivial to fix. I have to rewrite parts of the stackvar hand-
 *                   ling for this. 
 *       20.07.01:   Haven't cleaned up shit. Will add something like OpStrOff soon.
 *                   Seems to be a major pain to add tho --
 *       21.07.01:   It was not a pain -- dunno what kept me from doing this beforehand
 *                   Beautified the source a bit -- moved the member adding to a    
 *                   seperate function so I can more easily modify it -- this means
 *                   the maximum structure size will finally be enforced
 *                   Now that structure members are immediately set debugging is a lot
 *                   easier -- it is obvious now that the __cdecl recursion is far from
 *                   actually working
 */   
#include <ida.hpp>
#include <idp.hpp>
#include <loader.hpp>
#include <kernwin.hpp>
#include <frame.hpp>
#include <funcs.hpp> 
#include <xref.hpp>
#include <bytes.hpp>
#include <nalt.hpp>
#include <ua.hpp>
#include <struct.hpp>
#include <bytes.hpp>
#include <typeinf.hpp>
#include "insx86.hpp"   // home-grown documentation of cmd.itype, actually
                        // there is a "real" version along from ilfak, in
                        // the future might even use it ;>

#define  MAX_STRUC_SIZE 64000   // Maximum size of a structure in bytes.
                                // any member offset larger than this will
                                // be ignored (not implemented yet :)
#define  MAX_VTABLE_SIZE 0x1000
#define  FUNC_VISITED   0x8000
#define  MAX_STEPS      0x100

#define MODE_STACKVAR   0
#define MODE_REGISTER   1

#define RECUR_PUSH      1
#define RECUR_FASTCALL  0
#define RECUR_NONE      -1

int   __stdcall get_func_arg_size(ea_t func);
ea_t  __stdcall associate_push_to_call(ea_t ea_push);
ea_t  __stdcall stackvar_Op2(member_t *stck_var, ea_t ea_begin, ea_t ea_end);
int   __stdcall reg_opnd_num(int  reg, ea_t addr);
void  __stdcall ArgRec(ea_t ea_begin, ea_t ea_end, int ptr_mode, int recurse_mode, member_t *stck_var, reg_t reg_ptr, struc_t *structure);
ea_t  __stdcall GetPush(ea_t eaCall, int n);
ea_t  __stdcall register_xs(reg_t regnum, ea_t ea_begin, ea_t ea_end);
void  __stdcall ParseVtable(ea_t ea_begin, int ptr_mode, int recurse_mode, char *sz_var, reg_t reg_ptr, struc_t *structure);
reg_t __stdcall reg_string_to_num(char *regstring);

//--------------------------------------------------------------------------
//      initialize plugin

int __stdcall init(void)
{
   msg("...initializing BETA STRUCREC plugin...");

   if(strcmp(inf.procName, "metapc") != 0)
   {
      msg("  CPU Type should be \"metapc\" but is %s ! Aborting ...\n", inf.procName);
      return PLUGIN_SKIP;
   }
   msg("  ... init successful !\n");
   return PLUGIN_OK;
}

//--------------------------------------------------------------------------
//      terminate

void __stdcall term(void)
{
}

//--------------------------------------------------------------------------
//      The plugin method
char        StrucName[0x100];   // the structure name will be accessed by
                                // multiple functions, so its needed...

#define FALSE  0
#define TRUE   1
void __stdcall run(int arg)
{
   
   ea_t        func_start, func_end;  
   callui_t    ui_answer;
   struc_t     *new_struc;
   member_t    *stackvar;
   area_t      *fnc_area;
   int         ptrMode;       // Either 0 for stackvar or 1 for register
   int         recurseMode;   // Either 1 for PUSH recursion, 0 for FASTCALL recursion or -1 for no recursion
   bool        vtableMode;    // Either 0 for aborting ;-> or 1 for vtable parsing mode
   reg_t       reg_ptr;
   char        varname[200];  // local var name for vtable parsing   

   vtableMode = FALSE;        
   func_start = callui(ui_screenea).i32;
   func_end = -1;
   msg("%lx\n", func_start);
  
   if(isCode(getFlags(func_start)))
   {
      fnc_area = funcs.get_area(func_start);
 
      if((fnc_area < 0)||(func_start != fnc_area->startEA))
      {
         if(!callui(ui_askaddr, &func_end, "No function start found at current address, please enter end address:").cnd)
            return;  // if "CANCEL" is pushed, exit plugin
      }
      else
         func_end = fnc_area->endEA;
      if(func_end < func_start)
      {
         msg("[E] Error -- End address must not be lower than start address !\n");
         return;
      }
   }
   else
   {
      ui_answer = callui(ui_askyn, "Yes", "No", "Abort", 1, "No code found at current address ! Attempt vtable parsing ?");
      if(ui_answer.cnd != 1)
         return;  // Cancel it all
      else 
         vtableMode = TRUE;
   }
   
   ui_answer = callui(ui_askstr, 0, "arg_0", "Enter stackvar name or register pointer:");
      
   if(ui_answer.cptr == 0)
   {   
      msg("[E] Need pointer information, aborting ...\n");
      return;
   }

   stackvar = get_member_by_name(get_frame(func_start), ui_answer.cptr);
   reg_ptr = reg_string_to_num(ui_answer.cptr);

   if(stackvar)
      ptrMode = MODE_STACKVAR;
   else
      ptrMode = MODE_REGISTER;
      
   if((ptrMode == MODE_REGISTER)&&(reg_ptr == -1))
   {
      if(vtableMode)
         strcpy(varname, ui_answer.cptr);
      else
      {
         msg("[E] \"%s\" is not a valid stackvar name or register name...\n", ui_answer.cptr);
         return;
      }
   }

   ui_answer = callui(ui_askstr, 0, "unknstruc", "Please enter the reconstructed structure name:");
   strncpy(StrucName, ui_answer.cptr, sizeof(StrucName)-1);
        
   new_struc = get_struc(add_struc(BADADDR, StrucName, FALSE));
   if(new_struc == 0)
   {
      ui_answer = callui(ui_askyn, "New", "Existing", "Cancel", 1, "Work on existing structure or new one ?");
      if(ui_answer.cnd == -1)    // "Cancel" button was pushed
         return;
      if(ui_answer.cnd == 0)     // "Existing" button was pushed
         new_struc = get_struc(get_struc_id(StrucName));
      if(ui_answer.cnd == 1)     // "New" button was pushed
      {
         del_struc(get_struc(get_struc_id(StrucName)));
         new_struc = get_struc(add_struc(BADADDR, StrucName, FALSE));        
      }
   }

   recurseMode = (char)(callui(ui_askyn, "__cdecl", "__fastcall(notdone)", "None", -1, "Choose the recursion mode").cnd);
   
   msg("Ok, the settings are:\n");        
   msg("\t\tfunc_begin:\t%08lx\n", func_start);
   msg("\t\tfunc_end:\t%08lx\n", func_end);
   msg("\t\tptrMode:\t%08lx\n", ptrMode);
   msg("\t\trecurseMode:\t%08lx\n", recurseMode);
   msg("\t\tvtableMode:\t%08lx\n", vtableMode);

   if(vtableMode)
      ParseVtable(func_start, ptrMode, recurseMode, varname, reg_ptr, new_struc);
   else 
      ArgRec(func_start, func_end, ptrMode, recurseMode, stackvar, reg_ptr, new_struc);
   

}


bool    __stdcall op_to_struc_member(struc_t *structure, int offset, ea_t ea, int opnum)
{
   char  sz_memname[1000];
   typeinfo_t typinf;
      
   snprintf(sz_memname, 999, "%s_%lx", StrucName, offset);   
   msg("Adding structure member %s at offset %lx !\n", sz_memname, offset);
   add_struc_member(structure, sz_memname, offset, 0, &typinf, 1);
   
   if(!op_stroff(ea, opnum, &(structure->id), 4, 0))
   {
      msg("[!] op_stroff failed !\n ea=%lx opnum = %lx\n", ea, opnum);
      return(FALSE);
   }
   else
      return(TRUE);
}


reg_t   __stdcall reg_string_to_num(char *regstring)
{
   reg_t reg_ptr;
   if(stricmp(regstring, "eax") == 0)
      reg_ptr = r_eax;
   else if(stricmp(regstring, "ebx") == 0)
      reg_ptr = r_ebx;
   else if(stricmp(regstring, "ecx") == 0)
      reg_ptr = r_ecx;
   else if(stricmp(regstring, "edx") == 0)
      reg_ptr = r_edx;
   else if(stricmp(regstring, "esi") == 0)
      reg_ptr = r_esi;
   else if(stricmp(regstring, "edi") == 0)
      reg_ptr = r_edi;
   else if(stricmp(regstring, "ebp") == 0)
      reg_ptr = r_ebp;
   else if(stricmp(regstring, "esp") == 0)
      reg_ptr = r_esp;
   else  
      reg_ptr = (reg_t)-1;
   return(reg_ptr);
}
   


void __stdcall ParseVtable(ea_t ea_begin, int ptr_mode, int recurse_mode, char *sz_var, reg_t reg_ptr, struc_t *structure)
{
    ea_t    vtblstart, vtblend, vtblptr, funcptr;    
    area_t  *fnc_area;
    member_t *stck_var;
    
    vtblstart = ea_begin;
    vtblend = nextthat(ea_begin, ea_begin+MAX_VTABLE_SIZE, has_any_name);
    if(vtblend == BADADDR)
        return;
    msg("Allright, parsing vtable from %lx to %lx\n", vtblstart, vtblend);
    vtblptr = vtblstart;
    while(vtblptr < vtblend)
    {
        funcptr = get_long(vtblptr);
        if(funcptr == 0)
        {
            msg("NULL function pointer found, aborting !\n");
            break;
        }
        else
        {
            fnc_area = funcs.get_area(funcptr);
            stck_var = get_member_by_name(get_frame(funcptr), sz_var);
            msg("calling argrec with stackvar %lx (%s)", stck_var, sz_var);
            if(stck_var != 0)
               ptr_mode = MODE_STACKVAR;
            
            if((reg_ptr == -1)&&(stck_var == 0))
               msg("couldn't find either stackvar nor reg...\n");
            else if((fnc_area != (area_t *)BADADDR))
                ArgRec(fnc_area->startEA, fnc_area->endEA, ptr_mode, recurse_mode, stck_var, reg_ptr, structure);
            else
                msg("Error using functino at %lx with var %s !\n", funcptr, sz_var);    
        }
        vtblptr += 4;
    }
}

// ArgRec() is the main workhorse of the plugin. It starts tracing the function
// until the stackvar we follow is found in an instruction (Usually in a register
// load). Then it traces all accesses to this register until it is loaded with
// a different value and adds the offsets that get added to the register as offsets
// into the structure.

//    ea_begin       address at which to start parsing the code
//    ea_end         address up to which to parse the code
//    ptr_mode       either NULL for stackvar or 1 for register var mode
//    recurse_mode   either NULL for FASTCALL recursion, 1 for PUSH recursion or -1
//                   for no recursion
//    stck_var       pointer to the member_t on the stack if we're parsing stackvars
//    reg_ptr        reg_t number of the register to follow
//    structure      pointer to the structure we're creating

void __stdcall ArgRec(ea_t ea_begin, ea_t ea_end, int ptr_mode, int recurse_mode, member_t *stck_var, reg_t reg_ptr, struc_t *structure)
{
   ea_t        ea_curr_instr, ea_next_call;
   member_t    *new_stackvar;
   area_t      *fnc_area; // needed for __cdecl recursion
   reg_t       curr_reg;
   int         stackdiff; // used for calculating the right stack var

   ea_curr_instr = ea_begin;    

   if(ptr_mode == MODE_STACKVAR)        // STACKVAR pointer mode
   {
      while(ea_curr_instr <= ea_end)    // parse through the entire range
      {                                 // find first load of that stackvar
         ea_curr_instr = stackvar_Op2(stck_var, ea_curr_instr, ea_end); 
         if(ea_curr_instr == -1)        // if not found, return
            return;
         ua_ana0(ea_curr_instr);        // fill cmd structure
         curr_reg = (reg_t)cmd.Op1.reg; // figure out which register is loaded
         if(cmd.size != 0)              
            ea_curr_instr += cmd.size;  // trace to next instruction
         else
            ea_curr_instr++;            // ++ to prevent infinite looping
         
         ArgRec(ea_curr_instr, ea_end, MODE_REGISTER, recurse_mode, stck_var, curr_reg, structure);
         // recurse as register parser now
      }
   }
   else
   {
      curr_reg = reg_ptr;
      while(ea_curr_instr <= ea_end)
      { 
         // in case we have NO or PUSH recursion, get next register access
         // in case of fastcall recursion we have to trace all calls anyways
         if((recurse_mode == RECUR_PUSH)||(recurse_mode == RECUR_NONE))
            ea_curr_instr = register_xs(curr_reg, ea_curr_instr, ea_end);   
                // find next access to that register
         ua_ana0(ea_curr_instr);
                  
         if(ea_curr_instr == -1)
            return; // abort if no occurence found
         
         if(reg_opnd_num(curr_reg, ea_curr_instr) == 0)  //  first operand
         {
            if((cmd.itype == x86_movA)||    
               (cmd.itype == x86_movB)||
               (cmd.itype == x86_movsx)||
               (cmd.itype == x86_movzx)||
               (cmd.itype == x86_lea))
            {
               if(cmd.Op1.type == o_reg)
               {
                  if(cmd.Op2.reg == curr_reg)
                     op_to_struc_member(structure, cmd.Op2.addr, cmd.ea, 1);
                  break;
               }
               else if(cmd.Op1.type == o_phrase)
                  op_to_struc_member(structure, cmd.Op1.addr, cmd.ea, 0);
               else if(cmd.Op1.type == o_displ)
                  op_to_struc_member(structure, cmd.Op1.addr, cmd.ea, 0);
            }   // END of parsing mov* & lea
            else if(cmd.itype == x86_add)
            {
               op_to_struc_member(structure, cmd.Op2.value, cmd.ea, 1);
               break;   // the pointer has changed, forget it
            }
            else if(cmd.itype == x86_push)
            {
               if(cmd.Op1.type == o_reg)
               {
                  msg("[!] push found at %lx, need to recurse perhabs ?", ea_curr_instr);               
                  if(recurse_mode == RECUR_PUSH)
                  {
                    // we now need to find the call to which this push belongs, figure
                    // out the stack distance and recurse into it
                     if((ea_next_call = associate_push_to_call(ea_curr_instr)) != -1)
                     {
                        msg("stack at push: %lx\n", get_spd(get_func(ea_curr_instr), ea_curr_instr));                       
                        msg("call is: %lx\n", ea_next_call);
                        stackdiff = get_spd(get_func(ea_curr_instr), ea_curr_instr);
                        stackdiff -= get_spd(get_func(ea_next_call), ea_next_call);
                        stackdiff += 4; // adjust for return address
                        // we still have to insert code here for dealing with
                        // stub jumps and with infinite looping
                        fnc_area = funcs.get_area(get_first_cref_from(ea_next_call));
                        new_stackvar = get_member(get_frame(fnc_area->startEA), stackdiff);
                        msg("recursing into new call...\n");
                        ArgRec(fnc_area->startEA, fnc_area->endEA, MODE_STACKVAR, recurse_mode, new_stackvar, (reg_t)0, structure);
                     }
                  }
               }
               else
                  op_to_struc_member(structure, cmd.Op1.value, cmd.ea, 0);
            }
            else if((cmd.itype == x86_callA)||(cmd.itype == x86_callB)||(cmd.itype==x86_callC))
            {
               if(curr_reg == r_eax)
                  break;
            }
            else if((cmd.itype == x86_or)||(cmd.itype == x86_xor))
               break;   // reg was overwritten with new value, so forget it
            else
               op_to_struc_member(structure, cmd.Op1.addr, cmd.ea, 0);
         }
         if(reg_opnd_num(curr_reg, ea_curr_instr) == 1)  // second operand
         {
            if(cmd.Op2.type == o_reg)
            {
               if((cmd.itype == x86_movA)||
                  (cmd.itype == x86_movB))
               {
                  if(cmd.Op1.type == o_reg)  // recurse if mov found
                     ArgRec(ea_curr_instr+cmd.size, ea_end, 1, recurse_mode, stck_var, (reg_t)cmd.Op1.reg, structure);
                  else if(get_stkvar((cmd.Op1), cmd.Op1.addr, &(cmd.Op1.addr)) != 0)
                     ArgRec(ea_curr_instr, ea_end, 0, recurse_mode, get_stkvar((cmd.Op2), cmd.Op2.addr, &(cmd.Op2.addr)), (reg_t)0, structure);   
               }
            }
            if(cmd.Op2.value == 0)
               op_to_struc_member(structure, cmd.Op2.addr, cmd.ea, 1);
            else
               op_to_struc_member(structure, cmd.Op2.value, cmd.ea, 1);
         }
         if(reg_opnd_num(curr_reg, ea_curr_instr) == -1) // no operand (fastcall ?)
         {
            if((cmd.itype == x86_callA)||
               (cmd.itype == x86_callB)||
               (cmd.itype == x86_callC))
               msg("[!] call found, interesting for fastcall recusion\n");
            else
               msg("no accesss to our reg....\n");
         }
         if(cmd.size == 0)
            ea_curr_instr++;
         else
            ea_curr_instr += cmd.size;         
      }
   }
}

// This function will attempt to associate a PUSH with the associated call 
// instruction

ea_t  __stdcall associate_push_to_call(ea_t ea_push)
{
   ea_t  next_call, func;
   int   arg_num, steps = MAX_STEPS;
   int   stack_push;
   func_t   *function;
      
   next_call = ea_push;
   ua_ana0(ea_push);

   function = get_func(ea_push);       // Get the function frame structure
   if((function == NULL)||(function == (void *)-1))
      return(-1);
   stack_push = get_spd(function, ea_push);
   
   while(steps-- != 0)     // step forwards
   {
      if(get_spd(function, next_call) < stack_push)
         break;
      while((cmd.itype != x86_callA)&&(cmd.itype != x86_callB)&&(cmd.itype != x86_callC))
      {
         if(cmd.size == 0)
            next_call++;
         else
            next_call += cmd.size;
         ua_ana0(next_call);
      }  // trace until we find the next call       

      func = get_next_cref_from(next_call, get_first_cref_from(next_call)); 
      
      if((func == 0)||(func == -1))
         continue;
      
      ua_ana0(func); // now we need to deal with stub jmps
      if((cmd.itype == x86_jmpA)||(cmd.itype == x86_jmpB)||(cmd.itype == x86_jmpC)||(cmd.itype == x86_jmp_short))
         func = get_first_cref_from(func); 
      
      if((func == -1)||(func == 0))
         continue;
         
      arg_num = get_func_arg_size(func);
      if(arg_num == -1)
         continue;
      
      if((get_spd(function, next_call)-arg_num) <= (stack_push))
      {
         msg("returning %lx\n", next_call);
         return(next_call);
      }
      else
         continue;
   }
   return(-1);
}

// This function will return the estimated number of arguments for a particular
// function
int   __stdcall get_func_arg_size(ea_t func)
{
   struc_t  *func_frame;
   member_t *ret_addr; 
   int      arg_num;
      
   func_frame = get_frame(func);
   if(func_frame == 0)
      return(-1);
   ret_addr = get_member_by_name(func_frame, " r");
   if(ret_addr == 0)
      return(-1);
      
   arg_num = get_struc_last_offset(func_frame)-(ret_addr->soff);
   if(arg_num > 50)     // heuristic detection of skewed stackframes :)
      return(-1);
   return(arg_num);
}


// Returns the operand number (either 0 for first or 1 for second operand)
// in which the specified stackvar is found... returns -1 if not found at all

int   __stdcall stack_opnd_num(member_t *stck_var, ea_t addr)
{
   ua_ana0(addr);
   if(get_stkvar((cmd.Op2), cmd.Op2.addr, &(cmd.Op2.addr)) == stck_var)
      return(1);
   if(get_stkvar((cmd.Op1), cmd.Op1.addr, &(cmd.Op1.addr)) == stck_var)
      return(0);
      
   return(-1);
}

int   __stdcall reg_opnd_num(int  reg, ea_t addr)
{
   ua_ana0(addr);
   if(cmd.Op1.reg == reg)
      return(0);
   if(cmd.Op2.reg == reg)
      return(1);
   return(-1);
}

ea_t  __stdcall stackvar_Op2(member_t *stck_var, ea_t ea_begin, ea_t ea_end)
{
   while(ea_begin <= ea_end)
   {
      ua_ana0(ea_begin);   
      
      if((get_stkvar((cmd.Op2), cmd.Op2.addr, &(cmd.Op2.addr)) == stck_var)&&(cmd.Op2.type == o_displ))
      {
         if((cmd.Op2.reg == r_ebp)||(cmd.Op2.reg == r_esp))
            break;
      }
      if(cmd.size != 0)
         ea_begin += cmd.size;
      else
         ea_begin++;
   }
   return(ea_begin);
}

ea_t  __stdcall stackvar_Op1(member_t *stck_var, ea_t ea_begin, ea_t ea_end)
{
   while(ea_begin <= ea_end)
   {
      ua_ana0(ea_begin);   
      if(get_stkvar((cmd.Op1), cmd.Op1.addr, &(cmd.Op1.addr)) == stck_var)
         break;
      if(cmd.size != 0)
         ea_begin += cmd.size;
      else
         ea_begin++;
   }
   return(ea_begin);
}

ea_t  __stdcall stackvar_xs(member_t *stck_var, ea_t ea_begin, ea_t ea_end)
{
   while(ea_begin <= ea_end)
   {
      ua_ana0(ea_begin);   
      if(get_stkvar((cmd.Op1), cmd.Op1.addr, &(cmd.Op1.addr)) == stck_var)
         break;
      if(get_stkvar((cmd.Op2), cmd.Op2.addr, &(cmd.Op2.addr)) == stck_var)
         break;
      if(cmd.size != 0)
         ea_begin += cmd.size;
      else
         ea_begin++;
   }
   return(ea_begin);
}



ea_t  __stdcall register_read(reg_t regnum, ea_t ea_begin, ea_t ea_end)
{
   ua_ana0(ea_begin);
   ea_begin += cmd.size;
   
   while(cmd.Op2.reg != regnum)
   {
      ua_ana0(ea_begin);
      if(cmd.size == 0)
         ea_begin++;
      else
         ea_begin += cmd.size;        
   }     
   if(ea_begin >= ea_end)
      return(-1);
   else
      return(ea_begin-cmd.size);
}

ea_t  __stdcall register_write(reg_t regnum, ea_t ea_begin, ea_t ea_end)
{
   ua_ana0(ea_begin);
   ea_begin += cmd.size;

   while(cmd.Op1.reg != regnum)
   {
      ua_ana0(ea_begin);
      if(cmd.size == 0)
         ea_begin++;
      else
         ea_begin += cmd.size;        
   }     
   if(ea_begin >= ea_end)
      return(-1);
   else
      return(ea_begin-cmd.size);
}

ea_t  __stdcall register_xs(reg_t regnum, ea_t ea_begin, ea_t ea_end)
{
   while(ea_begin <= ea_end)
   {
      ua_ana0(ea_begin);
      if(((cmd.itype == x86_callA)||
         (cmd.itype == x86_callB)||
         (cmd.itype == x86_callC))&&(regnum == r_eax))
         break;      // in case EAX is followed and a call occurs
      
      if((cmd.Op1.type == o_reg)||(cmd.Op1.type == o_phrase)||(cmd.Op1.type == o_displ))
         if(cmd.Op1.reg == regnum)
            break;   
      if((cmd.Op2.type == o_reg)||(cmd.Op2.type == o_phrase)||(cmd.Op2.type == o_displ))
         if(cmd.Op2.reg == regnum)
            break;   
      if(cmd.size != 0)
         ea_begin += cmd.size;
      else
         ea_begin++;
   }
   return(ea_begin);  
}

/*  
void __stdcall DoRecursion(ea_t ea_func, member_t *stck_var, struc_t *structure)
{
   area_t   *fnc_area;
   ea_t     ea_trace_instr, ea_target_func, ea_push, ea_regload;
   struc_t  *fnc_frame;
   member_t *ret_addr, *push_var, *recurse_var;
   unsigned long  arg_size, maxsteps;
   uchar    regnum;
   
   fnc_area = funcs.get_area(ea_func);
   ea_trace_instr = fnc_area->startEA; // load ea_trace_instr with fnc begin
   
   while(ea_trace_instr < (fnc_area->endEA))
   {
      ua_ana0(ea_trace_instr);   // fill cmd structure
      if((cmd.itype == x86_callA)||(cmd.itype == x86_callB)||(cmd.itype ==x86_callC))
      {
         ea_target_func = get_next_cref_from(ea_trace_instr, get_first_cref_from(ea_trace_instr)); // get address of fnc
         
         ua_ana0(ea_target_func);
         if((cmd.itype == x86_jmpA)||(cmd.itype == x86_jmpB)||(cmd.itype == x86_jmpC)||(cmd.itype == x86_jmp_short))
            ea_target_func = get_first_cref_from(ea_target_func); // deal with stub jmps
         
         if(get_func(ea_target_func))
         {
            if(((get_func(ea_target_func)->flags) & FUNC_VISITED)==0)
            {
// Now the real fun begins... always remember: stackframes are just a special
// kind of structures          
               fnc_frame = get_frame(ea_target_func);
               ret_addr = get_member_by_name(fnc_frame, " r");
//         msg("looking at function %lx\n", ea_target_func);
               if(ret_addr != 0) // bail out if no retaddr can be found
               {
                  arg_size = get_struc_last_offset(fnc_frame)-(ret_addr->soff);
//            msg("arg_size for this function is %lx\n", arg_size);
// Being a nitpicker, I have to admit the above code line is inherently broken :)
// (get_struc_last_offset(fnc_frame)+get_member_size(get_member(fnc_frame, get_struc_last_offset(fnc_frame))))
// - ((ret_addr->soff)+4) would perhabs be better

                  if(arg_size < 50) // ignore functions with skewed stackframe
                  {
// since we now know how many arguments we have to the function we can start
// associating pushes with the call.               

                     for(arg_size ;arg_size > 0;arg_size -= 4)
                     {
                        if(arg_size > 50)
                           break;
                        ea_push = GetPush(ea_trace_instr, (arg_size/4));
                  
//                      msg("push found at %lx\n", ea_push);
                  
                        ua_ana0(ea_push);

                        if(cmd.Op1.type == o_reg)  // push simply pushing a register ?
                        {
// we will now need to trace back to where the register is loaded
                           regnum = cmd.Op1.reg;
                           ea_regload = get_first_cref_to(ea_push);
                           ua_ana0(ea_regload);
                           maxsteps = MAX_STEPS;
                           while((regnum != cmd.Op1.reg)&&(maxsteps >0))
                           {                             // register was loaded
//                            msg("Analyzing instuction %lx\n", ea_regload);
                              ea_regload = get_first_cref_to(ea_regload);
                              ua_ana0(ea_regload);
                              maxsteps--;
                           }
//                         msg("Register load for push found at %lx\n", ea_regload);
                           if((cmd.itype == x86_movA)||(cmd.itype == x86_movB))
                           {
                              push_var = get_stkvar(cmd.Op2, cmd.Op2.addr, &cmd.Op2.addr);
                              if(push_var == stck_var)   // we have a function to recurse into
                              {
//                           msg("arg_size right now is %lx, so arg %lx\n", arg_size, (arg_size/4));
                                 recurse_var = get_member(get_frame(ea_target_func), ((ret_addr->soff)+arg_size));
  
                                 msg("from %lx into %lx at addr %lx\n", ea_func, ea_target_func, ea_trace_instr);
// I know its bad style to simply use bits Ilfak has reserved, but well...
// if you come up with a better scheme to mark already visited funcs, go ahead :)                           
                                 get_func(ea_target_func)->flags |= FUNC_VISITED;
//                                 ArgRec(ea_target_func, recurse_var, structure); // recurse ! :)
                                 msg("returning from recursion...\n");
                              }
                           }
                        }
                        else if(cmd.Op1.type == o_phrase)   // push pushing something else ?
                        {
                           msg("---push is pushing o_phrase---not handled yet\n");
                        }
                     }
                  }
               }
            }
         }
      }
      ea_trace_instr += cmd.size;
      if(cmd.size == 0)
         ea_trace_instr++;    // break out of infinite loop in case size is 0
   }
}

/*
  This function uses IDA's built in stack tracking to find the n-th push. 
  We basically retrieve the stack pointer at the call and then trace regular code flow references
  backwards until we have found a place where the stack pointer had a 4*n bytes higher value.

  ea_t  GetPush(
          ea_t  eaCall,       // Address of the call
          int    n            // n -> Get 'n'th push before the call
          );

  Returns the address of the 'n'th push before the call          
*/

ea_t __stdcall GetPush(ea_t eaCall, int n)
{
   func_t   *func;
   ea_t     eaTrace;  
   int      SP;
   unsigned long  maxsteps;
   
   eaTrace = eaCall;
   func = get_func(eaTrace);       // Get the function frame structure
   SP = get_spd(func, eaTrace);    // Get the stack pointer at the call
   n *= 4;                        // multiply n by 4 to get the amount of stack correction we need
   maxsteps = MAX_STEPS;
   while((SP+n != get_spd(func, eaTrace))&&(eaTrace != -1)&&(maxsteps>0))
   {
      eaTrace = get_first_cref_to(eaTrace);  // Get first regular flow reference and parse back
      maxsteps--;
   }
   if(maxsteps == 0)
      eaTrace = -1;
      
   if(eaTrace == -1)
      msg("[!] Failed to find argument %ld to call at %lx\n", n, eaCall);
   return eaTrace;
}
        
        
        
                



//--------------------------------------------------------------------------
char comment[] = "StrucRec/metapc";
                  
char help[] =
        "This is the StrucRec/metapc plugin...\n"
        "You can use this plugin to reconstruct unknown\n"
        "structures from the disassembly. It will ask you\n"
        "for a ";
        
                
//--------------------------------------------------------------------------
// This is the preferred name of the plugin module in the menu system
// The preferred name may be overriden in plugins.cfg file

char wanted_name[] = "StrucRec/BETA/metapc";
                     
// This is the preferred hotkey for the plugin module
// The preferred hotkey may be overriden in plugins.cfg file
// Note: IDA won't tell you if the hotkey is not correct
//       It will just disable the hotkey.

char wanted_hotkey[] = "Alt-3";

//--------------------------------------------------------------------------
//
//      PLUGIN DESCRIPTION BLOCK
//
//--------------------------------------------------------------------------

extern "C" plugin_t PLUGIN = {
  IDP_INTERFACE_VERSION,
  0,                    // plugin flags
  init,                 // initialize

  term,                 // terminate. this pointer may be NULL.

  run,                  // invoke plugin

  comment,              // long comment about the plugin
                        // it could appear in the status line
                        // or as a hint

  help,                 // multiline help about the plugin

  wanted_name,          // the preferred short name of the plugin
  wanted_hotkey         // the preferred hotkey to run the plugin
};
